'use strict';

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol ? "symbol" : typeof obj; };

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _events = require('events');

var _events2 = _interopRequireDefault(_events);

var _clusterize = require('clusterize.js');

var _clusterize2 = _interopRequireDefault(_clusterize);

var _flattree = require('flattree');

var _lookupTable = require('./lookup-table');

var _lookupTable2 = _interopRequireDefault(_lookupTable);

var _renderer = require('./renderer');

var _polyfill = require('./polyfill');

var _utils = require('./utils');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { 'default': obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

var extend = function extend(target) {
    for (var _len = arguments.length, sources = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
        sources[_key - 1] = arguments[_key];
    }

    if (target === undefined || target === null) {
        throw new TypeError('Cannot convert undefined or null to object');
    }

    var output = Object(target);
    for (var index = 0; index < sources.length; index++) {
        var source = sources[index];
        if (source !== undefined && source !== null) {
            for (var key in source) {
                if (source.hasOwnProperty(key)) {
                    output[key] = source[key];
                }
            }
        }
    }
    return output;
};

var InfiniteTree = function (_events$EventEmitter) {
    _inherits(InfiniteTree, _events$EventEmitter);

    function InfiniteTree() {
        var options = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];

        _classCallCheck(this, InfiniteTree);

        // Assign options

        var _this = _possibleConstructorReturn(this, Object.getPrototypeOf(InfiniteTree).call(this));

        _this.options = {
            autoOpen: false,
            el: null,
            rowRenderer: _renderer.defaultRowRenderer
        };
        _this.state = {
            openNodes: [],
            rootNode: null,
            selectedNode: null
        };
        _this.clusterize = null;
        _this.tbl = new _lookupTable2['default']();
        _this.nodes = [];
        _this.rows = [];
        _this.scrollElement = null;
        _this.contentElement = null;

        _this.contentListener = function (evt) {
            var target = evt.target;
            var currentTarget = evt.currentTarget;


            (0, _polyfill.stopPropagation)(evt);

            if (target !== currentTarget) {
                var itemTarget = target;
                var handleToggler = false;

                while (itemTarget && itemTarget.parentElement !== currentTarget) {
                    if (itemTarget.className.indexOf('tree-toggler') >= 0) {
                        handleToggler = true;
                    }
                    itemTarget = itemTarget.parentElement;
                }

                var id = itemTarget.getAttribute('aria-id');
                var node = _this.getNodeById(id);

                if (!node) {
                    return;
                }

                // Click on the toggler to open/close a tree node
                if (handleToggler) {
                    _this.toggleNode(node);
                } else {
                    _this.selectNode(node);
                }
            }
        };

        _this.options = extend({}, _this.options, options);

        if (!_this.options.el) {
            console.error('Failed to initialize infinite-tree: el is not specified.', options);
            return _possibleConstructorReturn(_this);
        }

        _this.create();

        // Load tree data if it's provided
        if (options.data) {
            _this.loadData(options.data);
        }
        return _this;
    }

    _createClass(InfiniteTree, [{
        key: 'create',
        value: function create() {
            var _this2 = this;

            if (!this.options.el) {
                throw new Error('The element option is not specified.');
            }

            var scrollElement = document.createElement('div');
            scrollElement.className = (0, _utils.classNames)('infinite-tree', 'infinite-tree-scroll');
            var contentElement = document.createElement('div');
            contentElement.className = (0, _utils.classNames)('infinite-tree', 'infinite-tree-content');

            scrollElement.appendChild(contentElement);
            this.options.el.appendChild(scrollElement);

            this.clusterize = new _clusterize2['default']({
                tag: 'div',
                rows: [],
                scrollElem: scrollElement,
                contentElem: contentElement,
                no_data_class: 'infinite-tree-no-data',
                callbacks: {
                    // Will be called right before replacing previous cluster with new one.
                    clusterWillChange: function clusterWillChange() {},
                    // Will be called right after replacing previous cluster with new one.
                    clusterChanged: function clusterChanged() {},
                    // Will be called on scrolling. Returns progress position.
                    scrollingProgress: function scrollingProgress(progress) {
                        _this2.emit('scrollProgress', progress);
                    }
                }
            });

            this.scrollElement = scrollElement;
            this.contentElement = contentElement;

            (0, _polyfill.addEventListener)(this.contentElement, 'click', this.contentListener);
        }
    }, {
        key: 'clear',
        value: function clear() {
            this.clusterize.clear();
            this.tbl.clear();
            this.nodes = [];
            this.rows = [];
            this.state.openNodes = [];
            this.state.rootNode = null;
            this.state.selectedNode = null;
        }
    }, {
        key: 'destroy',
        value: function destroy() {
            (0, _polyfill.removeEventListener)(this.contentElement, 'click', this.contentListener);

            this.clear();

            if (this.clusterize) {
                this.clusterize.destroy(true); // True to remove all data from the list
                this.clusterize = null;
            }

            // Remove all child nodes
            while (this.contentElement.firstChild) {
                this.contentElement.removeChild(this.contentElement.firstChild);
            }
            while (this.scrollElement.firstChild) {
                this.scrollElement.removeChild(this.scrollElement.firstChild);
            }
            if (this.options.el) {
                var containerElement = this.options.el;
                while (containerElement.firstChild) {
                    containerElement.removeChild(containerElement.firstChild);
                }
            }
            this.contentElement = null;
            this.scrollElement = null;
        }
        // Updates list with new data

    }, {
        key: 'update',
        value: function update() {
            this.clusterize.update(this.rows);
        }
        // Inserts a new child node to a parent node at the specified index.
        // * If the parent is null or undefined, inserts the child at the specified index in the top-level.
        // * If the parent has children, the method adds the child to it at the specified index.
        // * If the parent does not have children, the method adds the child to the parent.
        // * If the index value is greater than or equal to the number of children in the parent, the method adds the child at the end of the children.
        // @param {object} newNode The object that defines the new child node.
        // @param {number} [index] The 0-based index of where to insert the child node. Defaults to 0.
        // @param {object} parentNode The object that defines the parent node.

    }, {
        key: 'addChildNodeAt',
        value: function addChildNodeAt(newNode, index, parentNode) {
            var _this3 = this;

            // Defaults to rootNode if parentNode is not specified
            parentNode = parentNode || this.state.rootNode;
            if (!(parentNode instanceof _flattree.Node)) {
                throw new Error('The parent node must be a Node object.');
            }

            var rowRenderer = this.options.rowRenderer;


            if (!newNode) {
                return false;
            }
            index = Number(index) || 0;
            if (index < 0) {
                index = 0;
            }

            // Inserts the new child at the specified index
            newNode.parent = parentNode;
            parentNode.children.splice(index, 0, newNode);

            var deleteCount = parentNode.state.total;

            // Update index
            index = parentNode.children.indexOf(newNode);

            var nodes = (0, _flattree.flatten)(parentNode.children, { openNodes: this.state.openNodes });

            // Update newNode
            newNode = parentNode.getChildAt(index);

            // Update nodes & rows
            var rows = nodes.map(function (node) {
                return rowRenderer(node);
            });
            var parentOffset = this.nodes.indexOf(parentNode);
            this.nodes.splice.apply(this.nodes, [parentOffset + 1, deleteCount].concat(nodes));
            this.rows.splice.apply(this.rows, [parentOffset + 1, deleteCount].concat(rows));

            // Update the lookup table with newly added nodes
            this.tbl.set(newNode.id, newNode);
            this.flatten(newNode).forEach(function (node) {
                if (node.id !== undefined) {
                    _this3.tbl.set(node.id, node);
                }
            });

            // Update the row corresponding to the parent node
            this.rows[parentOffset] = rowRenderer(parentNode);

            // Updates list with new data
            this.update();

            return true;
        }
        // Adds a node to the end of the list of children of a specified parent node.
        // * If the parent is null or undefined, inserts the child at the specified index in the top-level.
        // * If the parent has children, the method adds the child as the last child.
        // * If the parent does not have children, the method adds the child to the parent.
        // @param {object} newNode The object that defines the new child node.
        // @param {object} parentNode The object that defines the parent node.
        // @return {boolean} Returns true on success, false otherwise.

    }, {
        key: 'appendChildNode',
        value: function appendChildNode(newNode, parentNode) {
            // Defaults to rootNode if parentNode is not specified
            parentNode = parentNode || this.state.rootNode;
            if (!(parentNode instanceof _flattree.Node)) {
                throw new Error('The parent node must be a Node object.');
            }

            var index = parentNode.children.length;
            return this.addChildNodeAt(newNode, index, parentNode);
        }
        // Inserts the specified node after the reference node.
        // @param {object} newNode The object that defines the new sibling node.
        // @param {object} referenceNode The object that defines the current node.

    }, {
        key: 'insertNodeAfter',
        value: function insertNodeAfter(newNode, referenceNode) {
            if (!(referenceNode instanceof _flattree.Node)) {
                throw new Error('The reference node must be a Node object.');
            }

            var parentNode = referenceNode.getParent();
            var index = parentNode.children.indexOf(referenceNode) + 1;
            return this.addChildNodeAt(newNode, index, parentNode);
        }
        // Inserts the specified node before the reference node.
        // @param {object} newNode The object that defines the new sibling node.
        // @param {object} referenceNode The object that defines the current node.

    }, {
        key: 'insertNodeBefore',
        value: function insertNodeBefore(newNode, referenceNode) {
            if (!(referenceNode instanceof _flattree.Node)) {
                throw new Error('The reference node must be a Node object.');
            }

            var parentNode = referenceNode.getParent();
            var index = parentNode.children.indexOf(referenceNode);
            return this.addChildNodeAt(newNode, index, parentNode);
        }
        // Closes a node to hide its children.
        // @param {object} node The object that defines the node.
        // @return {boolean} Returns true on success, false otherwise.

    }, {
        key: 'closeNode',
        value: function closeNode(node) {
            var rowRenderer = this.options.rowRenderer;

            // Retrieve node index

            var nodeIndex = this.nodes.indexOf(node);
            if (nodeIndex < 0) {
                throw new Error('Invalid node specified: node.id=' + JSON.stringify(node.id));
            }

            // Check if the closeNode action can be performed
            if (this.state.openNodes.indexOf(node) < 0) {
                return false;
            }

            // Keep selected node unchanged if "node" is equal to "this.state.selectedNode"
            if (this.state.selectedNode && this.state.selectedNode !== node) {
                // Action:
                //   close "node.0.0"
                //
                // Tree:
                // [0] - node.0
                // [1]  - node.0.0      => next selected node (index=1, total=2)
                // [2]      node.0.0.0  => last selected node (index=2, total=0)
                // [3]      node.0.0.1
                // [4]    node.0.1
                var selectedIndex = this.nodes.indexOf(this.state.selectedNode);
                var rangeFrom = nodeIndex + 1;
                var rangeTo = nodeIndex + node.state.total;

                if (rangeFrom <= selectedIndex && selectedIndex <= rangeTo) {
                    this.selectNode(node);
                }
            }

            node.state.open = false; // Set node.state.open to false
            var openNodes = this.state.openNodes.filter(function (node) {
                return node.hasChildren() && node.state.open;
            });
            this.state.openNodes = openNodes;

            var deleteCount = node.state.total;

            {
                // Traversing up through ancestors to subtract node.state.total.
                var p = node;
                while (p) {
                    p.state.total = p.state.total - deleteCount;
                    p = p.parent;
                }
            }

            // Remove elements from an array
            this.nodes.splice(nodeIndex + 1, deleteCount);
            this.rows.splice(nodeIndex + 1, deleteCount);
            this.rows[nodeIndex] = rowRenderer(node);

            // Emit the 'closeNode' event
            this.emit('closeNode', node);

            // Updates list with new data
            this.update();

            return true;
        }
        // Gets a list of child nodes.
        // @param {object} [node] The object that defines the node. If null or undefined, returns a list of top level nodes.
        // @return {array} Returns an array of child nodes.

    }, {
        key: 'getChildNodes',
        value: function getChildNodes() {
            var node = arguments.length <= 0 || arguments[0] === undefined ? null : arguments[0];

            if (node) {
                return node.children || [];
            }
            node = this.nodes.length > 0 ? this.nodes[0] : null;
            while (node && node.parent !== null) {
                node = node.parent;
            }
            return node && node.children || [];
        }
        // Gets a node by its unique id. This assumes that you have given the nodes in the data a unique id.
        // @param {string|number} id An unique node id. A null value will be returned if the id doesn't match.
        // @return {object} Returns the node the matches the id, null otherwise.

    }, {
        key: 'getNodeById',
        value: function getNodeById(id) {
            var node = this.tbl.get(id);
            if (!node) {
                // Find the first node that matches the id
                node = this.nodes.filter(function (node) {
                    return node.id === id;
                })[0];
                if (!node) {
                    return null;
                }
                this.tbl.set(node.id, node);
            }
            return node;
        }
        // Gets the selected node.
        // @return {object} Returns the selected node, or null if not selected.

    }, {
        key: 'getSelectedNode',
        value: function getSelectedNode() {
            return this.state.selectedNode;
        }
        // Gets an array of open nodes.
        // @return {array} Returns an array of open nodes.

    }, {
        key: 'getOpenNodes',
        value: function getOpenNodes() {
            // returns a shallow copy of an array into a new array object.
            return this.state.openNodes.slice();
        }
        // Loads data in the tree.
        // @param {object|array} data The data is an object or array of objects that defines the node.

    }, {
        key: 'loadData',
        value: function loadData() {
            var _this4 = this;

            var data = arguments.length <= 0 || arguments[0] === undefined ? [] : arguments[0];
            var _options = this.options;
            var autoOpen = _options.autoOpen;
            var rowRenderer = _options.rowRenderer;


            this.nodes = (0, _flattree.flatten)(data, { openAllNodes: autoOpen });

            // Clear lookup table
            this.tbl.clear();

            this.state.openNodes = this.nodes.filter(function (node) {
                return node.hasChildren() && node.state.open;
            });
            this.state.rootNode = function () {
                var node = arguments.length <= 0 || arguments[0] === undefined ? null : arguments[0];

                // Finding the root node
                while (node && node.parent !== null) {
                    node = node.parent;
                }
                return node;
            }(this.nodes[0]);
            this.state.selectedNode = null;

            // Update the lookup table with newly added nodes
            this.flatten(this.state.rootNode).forEach(function (node) {
                if (node.id !== undefined) {
                    _this4.tbl.set(node.id, node);
                }
            });

            // Update rows
            this.rows = this.nodes.map(function (node) {
                return rowRenderer(node);
            });

            // Updates list with new data
            this.update();
        }
        // Opens a node to display its children.
        // @param {object} node The object that defines the node.
        // @return {boolean} Returns true on success, false otherwise.

    }, {
        key: 'openNode',
        value: function openNode(node) {
            var _this5 = this;

            var rowRenderer = this.options.rowRenderer;

            // Retrieve node index

            var nodeIndex = this.nodes.indexOf(node);
            if (nodeIndex < 0) {
                throw new Error('Invalid node specified: node.id=' + JSON.stringify(node.id));
            }

            // Check if the openNode action can be performed
            if (this.state.openNodes.indexOf(node) >= 0) {
                return false;
            }

            node.state.open = true; // Set node.state.open to true
            var openNodes = [node].concat(this.state.openNodes); // the most recently used items first
            this.state.openNodes = openNodes;

            var nodes = (0, _flattree.flatten)(node.children, { openNodes: this.state.openNodes });
            var rows = nodes.map(function (node) {
                return rowRenderer(node);
            });

            // Insert an array inside another array
            this.nodes.splice.apply(this.nodes, [nodeIndex + 1, 0].concat(nodes));
            this.rows.splice.apply(this.rows, [nodeIndex + 1, 0].concat(rows));
            this.rows[nodeIndex] = rowRenderer(node);

            // Add all child nodes to the lookup table if the first child does not exist in the lookup table
            if (nodes.length > 0 && !this.tbl.get(nodes[0])) {
                nodes.forEach(function (node) {
                    if (node.id !== undefined) {
                        _this5.tbl.set(node.id, node);
                    }
                });
            }

            // Emit the 'openNode' event
            this.emit('openNode', node);

            // Updates list with new data
            this.update();

            return true;
        }
        // Removes a node.
        // @param {object} node The object that defines the node.

    }, {
        key: 'removeNode',
        value: function removeNode(node) {}
        // TODO

        // Sets the current scroll position to this node.
        // @param {object} node The object that defines the node.
        // @return {number} Returns the vertical scroll position, or -1 on error.

    }, {
        key: 'scrollToNode',
        value: function scrollToNode(node) {
            // Retrieve node index
            var nodeIndex = this.nodes.indexOf(node);
            if (nodeIndex < 0) {
                return -1;
            }
            if (!this.contentElement) {
                return -1;
            }
            // Get the offset height of the first child element that contains the "tree-item" class.
            var firstChild = this.contentElement.querySelectorAll('.tree-item')[0];
            var rowHeight = firstChild && firstChild.offsetHeight || 0;
            return this.scrollTop(nodeIndex * rowHeight);
        }
        // Gets (or sets) the current vertical position of the scroll bar.
        // @param {number} [value] An integer that indicates the new position to set the scroll bar to.
        // @return {number} Returns the vertical scroll position.

    }, {
        key: 'scrollTop',
        value: function scrollTop(value) {
            if (!this.scrollElement) {
                return 0;
            }
            if (value !== undefined) {
                this.scrollElement.scrollTop = Number(value);
            }
            return this.scrollElement.scrollTop;
        }
        // Selects a node.
        // @param {object} node The object that defines the node. If null or undefined, deselects the current node.
        // @return {boolean} Returns true on success, false otherwise.

    }, {
        key: 'selectNode',
        value: function selectNode() {
            var node = arguments.length <= 0 || arguments[0] === undefined ? null : arguments[0];
            var rowRenderer = this.options.rowRenderer;


            if (node === null) {
                // Deselect the current node
                if (this.state.selectedNode) {
                    var selectedNode = this.state.selectedNode;
                    var selectedIndex = this.nodes.indexOf(selectedNode);

                    selectedNode.state.selected = false;
                    this.rows[selectedIndex] = rowRenderer(selectedNode);
                    this.state.selectedNode = null;

                    // Emit the 'selectNode' event
                    this.emit('selectNode', null);

                    // Updates list with new data
                    this.update();

                    return true;
                }

                return false;
            }

            // Retrieve node index
            var nodeIndex = this.nodes.indexOf(node);
            if (nodeIndex < 0) {
                throw new Error('Invalid node specified: node.id=' + JSON.stringify(node.id));
            }

            // Select this node
            if (this.state.selectedNode !== node) {
                node.state.selected = true;
                this.rows[nodeIndex] = rowRenderer(node);
            }

            // Deselect the current node
            if (this.state.selectedNode) {
                var _selectedNode = this.state.selectedNode;
                var _selectedIndex = this.nodes.indexOf(_selectedNode);
                _selectedNode.state.selected = false;
                this.rows[_selectedIndex] = rowRenderer(_selectedNode);
            }

            if (this.state.selectedNode !== node) {
                this.state.selectedNode = node;

                // Emit the 'selectNode' event
                this.emit('selectNode', node);
            } else {
                this.state.selectedNode = null;

                // Emit the 'selectNode' event
                this.emit('selectNode', null);
            }

            // Updates list with new data
            this.update();

            return true;
        }
        // Toggles a node to display or hide its children.
        // @param {object} node The object that defines the node.

    }, {
        key: 'toggleNode',
        value: function toggleNode(node) {
            if (this.state.openNodes.indexOf(node) >= 0) {
                // close node
                this.closeNode(node);
            } else {
                // open node
                this.openNode(node);
            }
        }
        // Serializes the current state of a node to a JSON string.
        // @param {object} node The object that defines the node. If null, returns the whole tree.

    }, {
        key: 'toString',
        value: function toString() {
            var node = arguments.length <= 0 || arguments[0] === undefined ? null : arguments[0];

            var traverse = function traverse(node) {
                var s = '[';
                if (node && node.children) {
                    var _loop = function _loop(i) {
                        var list = [];
                        s = s + '{';
                        Object.keys(node).forEach(function (key) {
                            var value = node[key];
                            if (key === 'parent') {
                                // ignore parent
                                return;
                            }
                            if (key === 'children') {
                                // traverse child nodes
                                list.push('"' + key + '":' + traverse(node.children[i]));
                                return;
                            }
                            if (typeof value === 'string' || (typeof value === 'undefined' ? 'undefined' : _typeof(value)) === 'object') {
                                list.push('"' + key + '":' + JSON.stringify(value));
                            } else {
                                // primitive types
                                list.push('"' + key + '":' + value);
                            }
                        });
                        s = s + list.join(',');
                        s = s + '}' + (i === node.children.length - 1 ? '' : ',');
                    };

                    for (var i = 0; i < node.children.length; ++i) {
                        _loop(i);
                    }
                }
                s = s + ']';
                return s;
            };

            if (!node) {
                node = this.state.rootNode;
            }

            return traverse(node);
        }
        // Flattens parent-child nodes by performing full tree traversal using child-parent link.
        // No recursion or stack is involved.
        // @param {object} parentNode The object that defines the parent node.
        // @return {array} Returns a flattened list of child nodes, not including the parent node.

    }, {
        key: 'flatten',
        value: function flatten(parentNode) {
            var list = [];

            if (parentNode === undefined) {
                parentNode = this.state.rootNode;
            }

            // Ignore parent node
            var node = parentNode.getFirstChild();
            while (node) {
                list.push(node);
                if (node.hasChildren()) {
                    node = node.getFirstChild();
                } else {
                    // find the parent level
                    while (node.getNextSibling() === null && node.parent !== parentNode) {
                        // use child-parent link to get to the parent level
                        node = node.getParent();
                    }

                    // Get next sibling
                    node = node.getNextSibling();
                }
            }

            return list;
        }
        // Updates the data of a node.
        // @param {object} node
        // @param {object} data The data object.

    }, {
        key: 'updateNode',
        value: function updateNode(node, data) {
            var rowRenderer = this.options.rowRenderer;

            // Retrieve node index

            var nodeIndex = this.nodes.indexOf(node);
            if (nodeIndex < 0) {
                throw new Error('Invalid node specified: node.id=' + JSON.stringify(node.id));
            }

            // The static attributes (i.e. children, parent, and state) are being protected
            var _node = node;
            var children = _node.children;
            var parent = _node.parent;
            var state = _node.state;

            node = extend(node, data, { children: children, parent: parent, state: state });

            this.rows[nodeIndex] = rowRenderer(node);

            // Updates list with new data
            this.update();
        }
    }]);

    return InfiniteTree;
}(_events2['default'].EventEmitter);

// IE8 compatibility output


module.exports = InfiniteTree;